using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using VSCom.CanApi;

namespace VSCom.J1939Api
{
    class VSCAN_J1939
    {
        public const string errHandleInvalid = "Handle is invalid (no Init done?)";
        /**
         * @brief API Version Structure
         */
        [StructLayout(LayoutKind.Sequential)]
        public struct VSCAN_J1939_VERSION
        {
            public byte Major;      /**< major number */
            public byte Minor;      /**< minor number */
            public byte SubMinor;   /**< subminor number */
        } 

        /**
         * @brief Message Structure
         */
        public struct VSCAN_J1939_MSG
        {
            public UInt32 PGN;    /**< Parameter Group Number */
            public byte[] Data;   /**< data bytes */
            public UInt16 Size;   /**< size of the data (max 1785 bytes / 7 * 255) */
            public byte Dst;      /**< destination address */
            public byte Src;      /**< our source address */
            public byte Priority; /**< message priority (max value of 7 / 0 = highest priority) */
            public byte Status;   /**< Tx: one of VSCAN_J1939_STATUS_... */
        } 
        [StructLayout(LayoutKind.Sequential)]
        public struct VSCAN_J1939_MSG_NATIVE
        {
            public UInt32 PGN;    /**< Parameter Group Number */
            public IntPtr Data;   /**< data bytes */
            public UInt16 Size;   /**< size of the data (max 1785 bytes / 7 * 255) */
            public byte Dst;      /**< destination address */
            public byte Src;      /**< our source address */
            public byte Priority; /**< message priority (max value of 7 / 0 = highest priority) */
            public byte Status;   /**< Tx: one of VSCAN_J1939_STATUS_... */
        } 

        /**
         * @brief Name Field Structure
         */
        [StructLayout(LayoutKind.Sequential)]
        public struct VSCAN_J1939_NAME
        {
            public UInt32 Id;           /**< Identity Number */
            public UInt16 ManCode;      /**< Manufacturer Code */
            public byte EcuInst;        /**< ECU (Electronic Control Unit) Instance */
            public byte FuncInst;       /**< Function Instance */
            public byte Func;           /**< Function */
            public byte VehicleSys;     /**< Vehicle System */
            public byte VehicleSysInst; /**< Vehicle System Instance */
            public byte IndustryGroup;  /**< Industry Group */
            public byte ArbAddrCapable; /**< Arbitrary Address Capable */
        }

        public delegate byte VSCAN_J1939_ADDR_CALLBACK();
        public delegate void VSCAN_J1939_PGN_CALLBACK(IntPtr Msg);

        /**
         * @name Error Constants
         * @{ */
        public const Int32 VSCAN_J1939_ERR_OK                   =  0;
        public const Int32 VSCAN_J1939_ERR_ERR					= -1;
        public const Int32 VSCAN_J1939_ERR_INVALID_HANDLE		= -2;
        public const Int32 VSCAN_J1939_ERR_INVALID_PARAMETER	= -3;
        public const Int32 VSCAN_J1939_ERR_MEMORY				= -4;
        public const Int32 VSCAN_J1939_ERR_SUBAPI				= -5;
        public const Int32 VSCAN_J1939_ERR_ADDR_UNAVAILABLE	    = -6;	/**< no addr from users callback function */
        public const Int32 VSCAN_J1939_ERR_ADDR_CLAIM			= -7;	/**< claiming our address failed */
        public const Int32 VSCAN_J1939_ERR_DATA_TO_SMALL		= -8;	/**< size must be >= 8 bytes */
        /** @} */

        /**
         * @name Sent Status Constants
         * @{ */
        public const byte VSCAN_J1939_STATUS_SENDING = 0x01;	/**< send is in progress */
        public const byte VSCAN_J1939_STATUS_SENT	 = 0x02;	/**< send was okay */
        public const byte VSCAN_J1939_STATUS_ERROR	 = 0x03;	/**< error during send */
        public const byte VSCAN_J1939_STATUS_FILL_IN = 0xff;	/**< use this status field during write */
        /** @} */

        /**
         * @name Default Constants
         * @{ */
        public const Int32 VSCAN_J1939_PGN_REQUEST		= 59904;
        public const Int32 VSCAN_J1939_PGN_ACK			= 59392;
        public const byte VSCAN_J1939_ADDR_BROADCAST	= 255;
        /** @} */

        [DllImport("vs_can_j1939", EntryPoint = "VSCAN_J1939_Init", CallingConvention = CallingConvention.Cdecl)]
        private static extern Int32 VSCAN_J1939_Init(Int32 Handle, VSCAN_J1939_NAME Name, VSCAN_J1939_ADDR_CALLBACK Callback);
        [DllImport("vs_can_j1939", EntryPoint = "VSCAN_J1939_Free", CallingConvention = CallingConvention.Cdecl)]
        private static extern Int32 VSCAN_J1939_Free(Int32 Handle);
        [DllImport("vs_can_j1939", EntryPoint = "VSCAN_J1939_Write", CallingConvention = CallingConvention.Cdecl)]
        private static extern Int32 VSCAN_J1939_Write(Int32 Handle, VSCAN_J1939_MSG_NATIVE[] Buf, UInt32 Size, ref UInt32 Written);
        [DllImport("vs_can_j1939", EntryPoint = "VSCAN_J1939_Request", CallingConvention = CallingConvention.Cdecl)]
        private static extern Int32 VSCAN_J1939_Request(Int32 Handle, UInt32 Pgn, byte Addr);
        [DllImport("vs_can_j1939", EntryPoint = "VSCAN_J1939_Read", CallingConvention = CallingConvention.Cdecl)]
        private static extern Int32 VSCAN_J1939_Read(Int32 Handle, IntPtr Buf, UInt32 Size, ref UInt32 Read);
        [DllImport("vs_can_j1939", EntryPoint = "VSCAN_J1939_SetPgnCallback", CallingConvention = CallingConvention.Cdecl)]
        private static extern Int32 VSCAN_J1939_SetPgnCallback(Int32 Handle, UInt32 Pgn, VSCAN_J1939_PGN_CALLBACK Callback);
        [DllImport("vs_can_j1939", EntryPoint = "VSCAN_J1939_Read_BufferCount", CallingConvention = CallingConvention.Cdecl)]
        private static extern UInt32 VSCAN_J1939_Read_BufferCount(Int32 Handle);
        [DllImport("vs_can_j1939", EntryPoint = "VSCAN_J1939_Read_BufferClear", CallingConvention = CallingConvention.Cdecl)]
        private static extern void VSCAN_J1939_Read_BufferClear(Int32 Handle);
        [DllImport("vs_can_j1939", EntryPoint = "VSCAN_J1939_ApiVersion", CallingConvention = CallingConvention.Cdecl)]
        private static extern VSCAN_J1939_VERSION VSCAN_J1939_ApiVersion();


        private Int32 handle = 0;

        public void Init(Int32 Handle, VSCAN_J1939_NAME Name, VSCAN_J1939_ADDR_CALLBACK Callback)
        {
            if (Handle <= 0)
                throw new VSCANException("Handle is invalid");
            handle = Handle;

            Int32 rc = VSCAN_J1939_Init(Handle, Name, Callback);
            if (rc != VSCAN_J1939_ERR_OK)
            {
                throw new VSCANException("Init failed " + rc);
            }
        }

        public void Free()
        {
            if (handle <= 0)
                throw new VSCANException(errHandleInvalid);

            Int32 rc = VSCAN_J1939_Free(handle);
            if (rc != VSCAN_J1939_ERR_OK)
            {
                throw new VSCANException("Free failed " + rc);
            }
            handle = 0;
        }

        public void Write(VSCAN_J1939_MSG[] Buf, UInt32 Size, ref UInt32 Written)
        {
            if (handle <= 0)
                throw new VSCANException(errHandleInvalid);

            VSCAN_J1939_MSG_NATIVE[] NBuf = new VSCAN_J1939_MSG_NATIVE[Size];

            for (int i = 0; i < Size; i++)
            {
                NBuf[i].PGN = Buf[i].PGN;
                NBuf[i].Src = Buf[i].Src;
                NBuf[i].Dst = Buf[i].Dst;
                NBuf[i].Priority = Buf[i].Priority;
                NBuf[i].Status = Buf[i].Status;
                if (Buf[i].Size == 0)
                    NBuf[i].Size = (UInt16)Buf[i].Data.Length;
                else
                    NBuf[i].Size = Buf[i].Size;
                NBuf[i].Data = Marshal.AllocHGlobal(Marshal.SizeOf(typeof(byte)) * NBuf[i].Size);
                Marshal.Copy(Buf[i].Data, 0, NBuf[i].Data, NBuf[i].Size);
            }
            
            Int32 rc = VSCAN_J1939_Write(handle, NBuf, Size, ref Written);

            for (int i = 0; i < Size; i++)
                Marshal.FreeHGlobal(NBuf[i].Data);
            
            if (rc != VSCAN_J1939_ERR_OK)
            {
                throw new VSCANException("Write failed " + rc);
            }
        }

        public void Request(UInt32 Pgn, byte Addr)
        {
            if (handle <= 0)
                throw new VSCANException(errHandleInvalid);

            Int32 rc = VSCAN_J1939_Request(handle, Pgn, Addr);

            if (rc != VSCAN_J1939_ERR_OK)
            {
                throw new VSCANException("Request failed " + rc);
            }
        }

        public void Read(ref VSCAN_J1939_MSG[] Buf, UInt32 Size, ref UInt32 Read)
        {
            if (handle <= 0)
                throw new VSCANException(errHandleInvalid);

            VSCAN_J1939_MSG_NATIVE[] NBuf = new VSCAN_J1939_MSG_NATIVE[Size];

            IntPtr pBuf = Marshal.AllocHGlobal(Marshal.SizeOf(NBuf[0]) * (int)Size);
            Marshal.StructureToPtr(NBuf[0], pBuf, false);

            Read = 0;
            Int32 rc = VSCAN_J1939_Read(handle, pBuf, Size, ref Read);

            IntPtr pCurrent = pBuf;
            for (int i = 0; i < Read; i++)
            {
                NBuf[i] = (VSCAN_J1939_MSG_NATIVE)Marshal.PtrToStructure(pCurrent, typeof(VSCAN_J1939_MSG_NATIVE));
                pCurrent = (IntPtr)((int)pCurrent + Marshal.SizeOf(NBuf[0]));
            }
            Marshal.FreeHGlobal(pBuf); 

            if (rc != VSCAN_J1939_ERR_OK)
            {
                throw new VSCANException("Read failed " + rc);
            }

            for (int i = 0; i < Size; i++)
            {
                Buf[i].PGN = NBuf[i].PGN;
                Buf[i].Src = NBuf[i].Src;
                Buf[i].Dst = NBuf[i].Dst;
                Buf[i].Priority = NBuf[i].Priority;
                Buf[i].Status = NBuf[i].Status;
                Buf[i].Size = NBuf[i].Size;
                Buf[i].Data = new byte[Buf[i].Size];
                Marshal.Copy((IntPtr)NBuf[i].Data, Buf[i].Data, 0, NBuf[i].Size);
            }
        }

        public void SetPgnCallback(UInt32 Pgn, VSCAN_J1939_PGN_CALLBACK Callback)
        {
            if (handle <= 0)
                throw new VSCANException(errHandleInvalid);

            Int32 rc = VSCAN_J1939_SetPgnCallback(handle, Pgn, Callback);

            if (rc != VSCAN_J1939_ERR_OK)
            {
                throw new VSCANException("SetPgnCallback failed " + rc);
            }
        }

        public UInt32 Read_BufferCount()
        {
            if (handle <= 0)
                throw new VSCANException(errHandleInvalid);

            return VSCAN_J1939_Read_BufferCount(handle);
        }

        public void Read_BufferClear()
        {
            if (handle <= 0)
                throw new VSCANException(errHandleInvalid);

            VSCAN_J1939_Read_BufferClear(handle);
        }

        public VSCAN_J1939_VERSION ApiVersion()
        {
            if (handle <= 0)
                throw new VSCANException(errHandleInvalid);

            return VSCAN_J1939_ApiVersion();
        }
    }
}
