/*!
 * \file
 *
 * \version
 * 1.0.0
 *
 * \brief
 * VSCAN J1939 API implementation (header file)
 *
 * This is the implementation of the VSCAN J1939 API.
 * The library is available for Windows and Linux.
 * 
 * Copyright (c) 2017 by VScom
 */
#ifndef _VS_CAN_J1939_H_
#define _VS_CAN_J1939_H_

#include "vs_can_api.h"

#ifdef WIN32
#include <windows.h>
#else
#include <semaphore.h>
typedef void VOID;
typedef char CHAR;
typedef unsigned long DWORD;
typedef unsigned char UCHAR;
typedef unsigned short USHORT;
typedef unsigned long ULONG;
#endif


#define VSCAN_J1939_ERR_OK					 0
#define VSCAN_J1939_ERR_ERR					-1
#define VSCAN_J1939_ERR_INVALID_HANDLE		-2
#define VSCAN_J1939_ERR_INVALID_PARAMETER	-3
#define VSCAN_J1939_ERR_MEMORY				-4
#define VSCAN_J1939_ERR_SUBAPI				-5
#define VSCAN_J1939_ERR_ADDR_UNAVAILABLE	-6	// no addr from users callback function
#define VSCAN_J1939_ERR_ADDR_CLAIM			-7	// claiming our address failed
#define VSCAN_J1939_ERR_DATA_TO_SMALL		-8	// size must be >= 8 bytes
#define VSCAN_J1939_ERR_UNSUPPORTED_DEVICE	-9	// this device is not for J1939


#define VSCAN_J1939_STATUS_SENDING	0x01	// send is in progress
#define VSCAN_J1939_STATUS_SENT		0x02	// send was okay
#define VSCAN_J1939_STATUS_ERROR	0x03	// error during send
#define VSCAN_J1939_STATUS_FILL_IN	0xff	// use this status field during write

#define VSCAN_J1939_PGN_REQUEST		59904
#define VSCAN_J1939_PGN_ACK			59392

#define VSCAN_J1939_ADDR_BROADCAST	255


typedef int	VSCAN_J1939_STATUS;
typedef UINT8	VSCAN_J1939_ADDR;
typedef UINT32	VSCAN_J1939_PGN;

// Version Info
typedef struct
{
	UINT8 Major;
	UINT8 Minor;
	UINT8 SubMinor;
} VSCAN_J1939_VERSION;

// Message Structure
typedef struct
{
   UINT32 PGN;				// Parameter Group Number
   UINT8 *Data;				// data which should be transmitted
   UINT16 Size;				// size of the data (max 1785 bytes / 7 * 255)
   UINT8 Dst;				// destination address
   UINT8 Src;				// our source address
   UINT8 Priority;			// message priority (max value of 7 / 0 = highest priority)
   UINT8 Status;			// Tx: one of VSCAN_J1939_STATUS_...
} VSCAN_J1939_MSG;

// Name Field Structure
typedef struct
{
	UINT32 Id;				// Identity Number
	UINT16 ManCode;			// Manufacturer Code
	UINT8 EcuInst;			// ECU (Electronic Control Unit) Instance
	UINT8 FuncInst;			// Function Instance
	UINT8 Func;				// Function
	UINT8 VehicleSys;		// Vehicle System
	UINT8 VehicleSysInst;	// Vehicle System Instance
	UINT8 IndustryGroup;	// Industry Group
	UINT8 ArbAddrCapable;	// Arbitrary Address Capable
} VSCAN_J1939_NAME;


#ifdef __cplusplus
extern "C" {
#endif

typedef VSCAN_J1939_ADDR (*VSCAN_J1939_ADDR_CALLBACK)();
typedef VOID (*VSCAN_J1939_PGN_CALLBACK)(VSCAN_HANDLE Handle, VSCAN_J1939_MSG *Msg);

VSCAN_J1939_STATUS VSCAN_J1939_Init(VSCAN_HANDLE Handle, VSCAN_J1939_NAME Name, VSCAN_J1939_ADDR_CALLBACK Callback);
VSCAN_J1939_STATUS VSCAN_J1939_Free(VSCAN_HANDLE Handle);
VSCAN_J1939_STATUS VSCAN_J1939_Write(VSCAN_HANDLE Handle, VSCAN_J1939_MSG *Buf, DWORD Size, DWORD *Written);
VSCAN_J1939_STATUS VSCAN_J1939_Request(VSCAN_HANDLE Handle, VSCAN_J1939_PGN Pgn, VSCAN_J1939_ADDR Addr);
VSCAN_J1939_STATUS VSCAN_J1939_Read(VSCAN_HANDLE Handle, VSCAN_J1939_MSG *Buf, DWORD Size, DWORD *Read);
VSCAN_J1939_STATUS VSCAN_J1939_SetPgnCallback(VSCAN_HANDLE Handle, VSCAN_J1939_PGN Pgn, VSCAN_J1939_PGN_CALLBACK Callback);
DWORD VSCAN_J1939_Read_BufferCount(VSCAN_HANDLE Handle);
VOID VSCAN_J1939_Read_BufferClear(VSCAN_HANDLE Handle);
VSCAN_J1939_VERSION VSCAN_J1939_ApiVersion();

#ifdef __cplusplus
}
#endif

#endif

